@@ -18,8 +18,8 @@ DBISTATE_DECLARE;
   #define croak_if_db_is_null()   if (!imp_dbh->db)   croak("imp_dbh->db is NULL at line %d in %s", __LINE__, __FILE__)
   #define croak_if_stmt_is_null() if (!imp_sth->stmt) croak("imp_sth->stmt is NULL at line %d in %s", __LINE__, __FILE__)
 #else
-  #define croak_if_db_is_null() 
-  #define croak_if_stmt_is_null() 
+  #define croak_if_db_is_null()
+  #define croak_if_stmt_is_null()
 #endif
 
 
@@ -142,10 +142,9 @@ static int
 sqlite_is_number(pTHX_ const char *v)
 {
     const char *z = v;
-    int i, neg;
+    int neg;
     int digit = 0;
     int precision = 0;
-    double f;
     char str[30], format[10];
 
     if      (*z == '-') { neg = 1; z++; }
@@ -273,6 +272,11 @@ sqlite_db_commit(SV *dbh, imp_dbh_t *imp_dbh)
     dTHX;
     int rc;
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to commit on inactive database handle");
+        return FALSE;
+    }
+
     if (DBIc_is(imp_dbh, DBIcf_AutoCommit)) {
         /* We don't need to warn, because the DBI layer will do it for us */
         return TRUE;
@@ -303,6 +307,11 @@ sqlite_db_rollback(SV *dbh, imp_dbh_t *imp_dbh)
     dTHX;
     int rc;
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to rollback on inactive database handle");
+        return FALSE;
+    }
+
     if (DBIc_is(imp_dbh, DBIcf_BegunWork)) {
         DBIc_off(imp_dbh, DBIcf_BegunWork);
         DBIc_on(imp_dbh,  DBIcf_AutoCommit);
@@ -329,11 +338,11 @@ sqlite_db_disconnect(SV *dbh, imp_dbh_t *imp_dbh)
     dTHX;
     int rc;
     sqlite3_stmt *pStmt;
-    DBIc_ACTIVE_off(imp_dbh);
 
     if (DBIc_is(imp_dbh, DBIcf_AutoCommit) == FALSE) {
         sqlite_db_rollback(dbh, imp_dbh);
     }
+    DBIc_ACTIVE_off(imp_dbh);
 
 #if 0
     /*
@@ -371,7 +380,6 @@ sqlite_db_disconnect(SV *dbh, imp_dbh_t *imp_dbh)
             sqlite_error(dbh, rc, sqlite3_errmsg(imp_dbh->db));
         }
     }
-    imp_dbh->db = NULL;
 
     av_undef(imp_dbh->functions);
     SvREFCNT_dec(imp_dbh->functions);
@@ -395,6 +403,8 @@ sqlite_db_destroy(SV *dbh, imp_dbh_t *imp_dbh)
     if (DBIc_ACTIVE(imp_dbh)) {
         sqlite_db_disconnect(dbh, imp_dbh);
     }
+    imp_dbh->db = NULL;
+
     DBIc_IMPSET_off(imp_dbh);
 }
 
@@ -410,7 +420,7 @@ sqlite_db_STORE_attrib(SV *dbh, imp_dbh_t *imp_dbh, SV *keysv, SV *valuesv)
     if (strEQ(key, "AutoCommit")) {
         if (SvTRUE(valuesv)) {
             /* commit tran? */
-            if ( (!DBIc_is(imp_dbh, DBIcf_AutoCommit)) && (!sqlite3_get_autocommit(imp_dbh->db)) ) {
+            if ( DBIc_ACTIVE(imp_dbh) && (!DBIc_is(imp_dbh, DBIcf_AutoCommit)) && (!sqlite3_get_autocommit(imp_dbh->db)) ) {
                 sqlite_trace(dbh, imp_dbh, 3, "COMMIT TRAN");
                 rc = sqlite_exec(dbh, "COMMIT TRANSACTION");
                 if (rc != SQLITE_OK) {
@@ -439,7 +449,8 @@ sqlite_db_STORE_attrib(SV *dbh, imp_dbh_t *imp_dbh, SV *keysv, SV *valuesv)
         return TRUE;
     }
     if (strEQ(key, "unicode")) {
-        warn("\"unicode\" attribute will be deprecated. Use \"sqlite_unicode\" instead.");
+        if (DBIc_has(imp_dbh, DBIcf_WARN))
+            warn("\"unicode\" attribute will be deprecated. Use \"sqlite_unicode\" instead.");
 #if PERL_UNICODE_DOES_NOT_WORK_WELL
         sqlite_trace(dbh, imp_dbh, 3, form("Unicode support is disabled for this version of perl."));
         imp_dbh->unicode = 0;
@@ -475,7 +486,8 @@ sqlite_db_FETCH_attrib(SV *dbh, imp_dbh_t *imp_dbh, SV *keysv)
 #endif
    }
    if (strEQ(key, "unicode")) {
-        warn("\"unicode\" attribute will be deprecated. Use \"sqlite_unicode\" instead.");
+        if (DBIc_has(imp_dbh, DBIcf_WARN))
+            warn("\"unicode\" attribute will be deprecated. Use \"sqlite_unicode\" instead.");
 #if PERL_UNICODE_DOES_NOT_WORK_WELL
        sqlite_trace(dbh, imp_dbh, 3, "Unicode support is disabled for this version of perl.");
        return sv_2mortal(newSViv(0));
@@ -492,314 +504,16 @@ sqlite_db_last_insert_id(SV *dbh, imp_dbh_t *imp_dbh, SV *catalog, SV *schema, S
 {
     dTHX;
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to get last inserted id on inactive database handle");
+        return FALSE;
+    }
+
     croak_if_db_is_null();
 
     return sv_2mortal(newSViv((IV)sqlite3_last_insert_rowid(imp_dbh->db)));
 }
 
-/* ======================================================================
- * EXPERIMENTAL bindings for FTS3 TOKENIZERS
- * ====================================================================== */
-
-typedef struct perl_tokenizer {  
-  sqlite3_tokenizer base;
-  SV *coderef;                 /* the perl tokenizer is a coderef that takes
-                                  a string and returns a cursor coderef */
-} perl_tokenizer;
-
-typedef struct perl_tokenizer_cursor {
-  sqlite3_tokenizer_cursor base;
-  SV *coderef;                 /* ref to the closure that returns terms */
-  char *pToken;                /* storage for a copy of the last token */
-  int nTokenAllocated;         /* space allocated to pToken buffer */
-
-  /* members below are only used if the input string is in utf8 */
-  const char *pInput;          /* input we are tokenizing */
-  const char *lastByteOffset;  /* offset into pInput */
-  int lastCharOffset;          /* char offset corresponding to lastByteOffset */
-
-} perl_tokenizer_cursor;
-
-
-/*
-** Create a new tokenizer instance.
-** Will be called whenever a FTS3 table is created with 
-**   CREATE .. USING fts3( ... , tokenize=perl qualified::function::name)
-** where qualified::function::name is a fully qualified perl function
-*/
-static int perl_tokenizer_Create(
-  int argc, const char * const *argv,
-  sqlite3_tokenizer **ppTokenizer
-){
-  dTHX;
-  dSP;
-  int n_retval;
-  SV *retval;
-
-  perl_tokenizer *t;
-  t = (perl_tokenizer *) sqlite3_malloc(sizeof(*t));
-  if( t==NULL ) return SQLITE_NOMEM;
-  memset(t, 0, sizeof(*t));
-
-  ENTER;
-  SAVETMPS;
-
-  /* call the qualified::function::name */
-  PUSHMARK(SP);
-  PUTBACK;
-  n_retval = call_pv(argv[0], G_SCALAR);
-  SPAGAIN;
-
-  /* store a copy of the returned coderef into the tokenizer structure */
-  if (n_retval != 1) {
-    warn("tokenizer_Create returned %d arguments", n_retval);
-  }
-  retval = POPs;
-  t->coderef   = newSVsv(retval);
-  *ppTokenizer = &t->base;
-
-  PUTBACK;
-  FREETMPS;
-  LEAVE;
-
-  return SQLITE_OK;
-}
-
-
-/*
-** Destroy a tokenizer
-*/
-static int perl_tokenizer_Destroy(sqlite3_tokenizer *pTokenizer){
-  dTHX;
-  perl_tokenizer *t = (perl_tokenizer *) pTokenizer;
-  sv_free(t->coderef);
-  sqlite3_free(t);
-  return SQLITE_OK;
-}
-
-
-/*
-** Prepare to begin tokenizing a particular string.  The input
-** string to be tokenized is supposed to be pInput[0..nBytes-1] .. 
-** except that nBytes passed by fts3 is -1 (don't know why) ! 
-** This is passed to the tokenizer instance, which then returns a 
-** closure implementing the cursor (so the cursor is again a coderef).
-*/
-static int perl_tokenizer_Open(
-  sqlite3_tokenizer *pTokenizer,       /* Tokenizer object */
-  const char *pInput, int nBytes,      /* Input buffer */
-  sqlite3_tokenizer_cursor **ppCursor  /* OUT: Created tokenizer cursor */
-){
-  dTHX;
-  dSP;
-  U32 flags;
-  SV *perl_string;
-  int n_retval;
-
-  perl_tokenizer *t = (perl_tokenizer *)pTokenizer;
-
-  /* allocate and initialize the cursor struct */
-  perl_tokenizer_cursor *c;
-  c = (perl_tokenizer_cursor *) sqlite3_malloc(sizeof(*c));
-  memset(c, 0, sizeof(*c));
-  *ppCursor = &c->base;
-
-  /* flags for creating the Perl SV containing the input string */
-  flags = SVs_TEMP; /* will call sv_2mortal */
-
-  /* special handling if working with utf8 strings */
-  if (last_executed_dbh->unicode) { /* global var ... no better way ! */
-
-    /* data to keep track of byte offsets */
-    c->lastByteOffset = c->pInput = pInput;
-    c->lastCharOffset = 0;
-
-    /* string passed to Perl needs to be flagged as utf8 */
-    flags |= SVf_UTF8;
-  }
-
-  ENTER;
-  SAVETMPS;
-
-  /* build a Perl copy of the input string */
-  if (nBytes < 0) { /* we get -1 from fts3. Don't know why ! */
-    nBytes = strlen(pInput);
-  } 
-  perl_string = newSVpvn_flags(pInput, nBytes, flags); 
-
-  /* call the tokenizer coderef */
-  PUSHMARK(SP);
-  XPUSHs(perl_string);
-  PUTBACK;
-  n_retval = call_sv(t->coderef, G_SCALAR);
-  SPAGAIN;
-
-  /* store the cursor coderef returned by the tokenizer */
-  if (n_retval != 1) {
-    warn("tokenizer returned %d arguments", n_retval);
-  }
-  c->coderef = newSVsv(POPs);
-
-  PUTBACK;
-  FREETMPS;
-  LEAVE;
-  return SQLITE_OK;
-}
-
-/*
-** Close a tokenization cursor previously opened by a call to
-** perl_tokenizer_Open() above.
-*/
-static int perl_tokenizer_Close(sqlite3_tokenizer_cursor *pCursor){
-  perl_tokenizer_cursor *c = (perl_tokenizer_cursor *) pCursor;
-
-  dTHX;
-  sv_free(c->coderef);
-  sqlite3_free(c);
-  return SQLITE_OK;
-}
-
-/*
-** Extract the next token from a tokenization cursor.  The cursor must
-** have been opened by a prior call to perl_tokenizer_Open().
-*/
-static int perl_tokenizer_Next(
-  sqlite3_tokenizer_cursor *pCursor,  /* Cursor returned by perl_tokenizer_Open */
-  const char **ppToken,               /* OUT: *ppToken is the token text */
-  int *pnBytes,                       /* OUT: Number of bytes in token */
-  int *piStartOffset,                 /* OUT: Starting offset of token */
-  int *piEndOffset,                   /* OUT: Ending offset of token */
-  int *piPosition                     /* OUT: Position integer of token */
-){
-  perl_tokenizer_cursor *c = (perl_tokenizer_cursor *) pCursor;
-  int result;
-  int n_retval;
-  STRLEN n_a;
-  char *token;
-  char *byteOffset;
-  I32 hop;
-
-  dTHX;
-  dSP;
-
-  ENTER;
-  SAVETMPS;
-
-  /* call the cursor */
-  PUSHMARK(SP);
-  PUTBACK;
-  n_retval = call_sv(c->coderef, G_ARRAY);
-  SPAGAIN;
-
-  /* if we get back an empty list, there is no more token */
-  if (n_retval == 0) { 
-    result = SQLITE_DONE;
-  }
-  /* otherwise, get token details from the return list */
-  else {
-    if (n_retval != 5) {
-      warn("tokenizer cursor returned %d arguments", n_retval);
-    }
-    *piPosition    = POPi;
-    *piEndOffset   = POPi;
-    *piStartOffset = POPi;
-    *pnBytes       = POPi;
-    token          = POPpx;
-
-    if (c->pInput) { /* if working with utf8 data */
-
-      /* recompute *pnBytes in bytes, not in chars */
-      *pnBytes = strlen(token); 
-
-      /* recompute start/end offsets in bytes, not in chars */
-                   hop = *piStartOffset - c->lastCharOffset;
-            byteOffset = utf8_hop(c->lastByteOffset, hop);
-                   hop = *piEndOffset - *piStartOffset;                
-        *piStartOffset = byteOffset - c->pInput;
-            byteOffset = utf8_hop(byteOffset, hop);
-          *piEndOffset = byteOffset - c->pInput;                                  
-      /* remember where we are for next round */
-      c->lastCharOffset = *piEndOffset,
-      c->lastByteOffset = byteOffset;
-    }
-
-    /* make sure we have enough storage for copying the token */
-    if (*pnBytes > c->nTokenAllocated ){
-      char *pNew;
-      c->nTokenAllocated = *pnBytes + 20;
-      pNew = sqlite3_realloc(c->pToken, c->nTokenAllocated);
-      if( !pNew ) return SQLITE_NOMEM;
-      c->pToken = pNew;
-    }
-
-    /* need to copy the token into the C cursor before perl frees that
-       memory */
-    memcpy(c->pToken, token, *pnBytes);
-    *ppToken  = c->pToken;
-
-    result = SQLITE_OK;
-  }
-
-  PUTBACK;
-  FREETMPS;
-  LEAVE;
-
-  return result;
-}
-
-
-/*
-** The set of routines that implement the perl tokenizer
-*/
-sqlite3_tokenizer_module perl_tokenizer_Module = {
-  0,
-  perl_tokenizer_Create,
-  perl_tokenizer_Destroy,
-  perl_tokenizer_Open,
-  perl_tokenizer_Close,
-  perl_tokenizer_Next
-};
-
-
-/*
-** Register the perl tokenizer with FTS3
-*/
-int sqlite_db_register_fts3_perl_tokenizer(pTHX_ SV *dbh)
-{
-  D_imp_dbh(dbh);
-
-  int rc;
-  sqlite3_stmt *pStmt;
-  const char zSql[] = "SELECT fts3_tokenizer(?, ?)";
-  sqlite3_tokenizer_module *p = &perl_tokenizer_Module;
-
-  rc = sqlite3_prepare_v2(imp_dbh->db, zSql, -1, &pStmt, 0);
-  if( rc!=SQLITE_OK ){
-    return rc;
-  }
-
-  sqlite3_bind_text(pStmt, 1, "perl", -1, SQLITE_STATIC);
-  sqlite3_bind_blob(pStmt, 2, &p, sizeof(p), SQLITE_STATIC);
-  sqlite3_step(pStmt);
-
-  return sqlite3_finalize(pStmt);
-}
-
-
-/* ======================================================================
- * END # EXPERIMENTAL bindings for FTS3 TOKENIZERS
- * ====================================================================== */
-
-
-
-
-
-
-
-
-
-
-
 int
 sqlite_st_prepare(SV *sth, imp_sth_t *imp_sth, char *statement, SV *attribs)
 {
@@ -809,14 +523,14 @@ sqlite_st_prepare(SV *sth, imp_sth_t *imp_sth, char *statement, SV *attribs)
     D_imp_dbh_from_sth;
 
     if (!DBIc_ACTIVE(imp_dbh)) {
-      sqlite_error(sth, -2, "attempt to prepare on inactive database handle");
-      return FALSE; /* -> undef in lib/DBD/SQLite.pm */
+        sqlite_error(sth, -2, "attempt to prepare on inactive database handle");
+        return FALSE; /* -> undef in lib/DBD/SQLite.pm */
     }
 
 #if 0
     if (*statement == '\0') {
-      sqlite_error(sth, -2, "attempt to prepare empty statement");
-      return FALSE; /* -> undef in lib/DBD/SQLite.pm */
+        sqlite_error(sth, -2, "attempt to prepare empty statement");
+        return FALSE; /* -> undef in lib/DBD/SQLite.pm */
     }
 #endif
 
@@ -923,7 +637,9 @@ sqlite_st_execute(SV *sth, imp_sth_t *imp_sth)
         else {
             STRLEN len;
             const char *data;
+#if 0
             int numtype;
+#endif
             if (imp_dbh->unicode) {
                 sv_utf8_upgrade(value);
             }
@@ -1059,6 +775,11 @@ sqlite_st_fetch(SV *sth, imp_sth_t *imp_sth)
     int chopBlanks = DBIc_is(imp_sth, DBIcf_ChopBlanks);
     int i;
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(sth, -2, "attempt to fetch on inactive database handle");
+        return FALSE;
+    }
+
     croak_if_db_is_null();
     croak_if_stmt_is_null();
 
@@ -1418,6 +1139,10 @@ sqlite_db_busy_timeout(pTHX_ SV *dbh, int timeout )
 
     if (timeout) {
         imp_dbh->timeout = timeout;
+        if (!DBIc_ACTIVE(imp_dbh)) {
+            sqlite_error(dbh, -2, "attempt to set busy timeout on inactive database handle");
+            return -2;
+        }
         sqlite3_busy_timeout(imp_dbh->db, timeout);
     }
     return imp_dbh->timeout;
@@ -1515,9 +1240,15 @@ sqlite_db_create_function(pTHX_ SV *dbh, const char *name, int argc, SV *func)
 {
     D_imp_dbh(dbh);
     int rc;
+    SV *func_sv;
+
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to create function on inactive database handle");
+        return FALSE;
+    }
 
     /* Copy the function reference */
-    SV *func_sv = newSVsv(func);
+    func_sv = newSVsv(func);
     av_push( imp_dbh->functions, func_sv );
 
     croak_if_db_is_null();
@@ -1526,7 +1257,7 @@ sqlite_db_create_function(pTHX_ SV *dbh, const char *name, int argc, SV *func)
     rc = sqlite3_create_function( imp_dbh->db, name, argc, SQLITE_UTF8,
                                   func_sv,
                                   imp_dbh->unicode ? sqlite_db_func_dispatcher_unicode
-                                                   : sqlite_db_func_dispatcher_no_unicode, 
+                                                   : sqlite_db_func_dispatcher_no_unicode,
                                   NULL, NULL );
     if ( rc != SQLITE_OK ) {
         sqlite_error(dbh, rc, form("sqlite_create_function failed with error %s", sqlite3_errmsg(imp_dbh->db)));
@@ -1541,6 +1272,11 @@ sqlite_db_enable_load_extension(pTHX_ SV *dbh, int onoff)
     D_imp_dbh(dbh);
     int rc;
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to enable load extension on inactive database handle");
+        return FALSE;
+    }
+
     croak_if_db_is_null();
 
     /* COMPAT: sqlite3_enable_load_extension is only available for 3003006 or newer */
@@ -1585,7 +1321,7 @@ sqlite_db_aggr_new_dispatcher(pTHX_ sqlite3_context *context, aggrInfo *aggr_inf
     } else if ( count != 1 ) {
         int i;
 
-        aggr_info->err = newSVpvf("new() should return one value, got %d", 
+        aggr_info->err = newSVpvf("new() should return one value, got %d",
                                    count );
         /* Clear the stack */
         for ( i=0; i < count; i++ ) {
@@ -1629,7 +1365,7 @@ sqlite_db_aggr_step_dispatcher(sqlite3_context *context,
         sqlite_db_aggr_new_dispatcher(aTHX_ context, aggr);
     }
 
-    if ( aggr->err || !aggr->aggr_inst ) 
+    if ( aggr->err || !aggr->aggr_inst )
         goto cleanup;
 
     PUSHMARK(SP);
@@ -1694,7 +1430,7 @@ sqlite_db_aggr_finalize_dispatcher( sqlite3_context *context )
         aggr->aggr_inst = NULL;
         aggr->err = NULL;
         sqlite_db_aggr_new_dispatcher(aTHX_ context, aggr);
-    } 
+    }
 
     if  ( ! aggr->err && aggr->aggr_inst ) {
         PUSHMARK(SP);
@@ -1745,9 +1481,15 @@ sqlite_db_create_aggregate(pTHX_ SV *dbh, const char *name, int argc, SV *aggr_p
 {
     D_imp_dbh(dbh);
     int rc;
+    SV *aggr_pkg_copy;
+
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to create aggregate on inactive database handle");
+        return FALSE;
+    }
 
     /* Copy the aggregate reference */
-    SV *aggr_pkg_copy = newSVsv(aggr_pkg);
+    aggr_pkg_copy = newSVsv(aggr_pkg);
     av_push( imp_dbh->aggregates, aggr_pkg_copy );
 
     croak_if_db_is_null();
@@ -1755,7 +1497,7 @@ sqlite_db_create_aggregate(pTHX_ SV *dbh, const char *name, int argc, SV *aggr_p
     rc = sqlite3_create_function( imp_dbh->db, name, argc, SQLITE_UTF8,
                                   aggr_pkg_copy,
                                   NULL,
-                                  sqlite_db_aggr_step_dispatcher, 
+                                  sqlite_db_aggr_step_dispatcher,
                                   sqlite_db_aggr_finalize_dispatcher
                                 );
 
@@ -1841,6 +1583,11 @@ sqlite_db_create_collation(pTHX_ SV *dbh, const char *name, SV *func)
 
     SV *func_sv = newSVsv(func);
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to create collation on inactive database handle");
+        return FALSE;
+    }
+
     croak_if_db_is_null();
 
     /* Check that this is a proper collation function */
@@ -1858,10 +1605,10 @@ sqlite_db_create_collation(pTHX_ SV *dbh, const char *name, SV *func)
     av_push( imp_dbh->functions, func_sv );
 
     /* Register the func within sqlite3 */
-    rv = sqlite3_create_collation( 
+    rv = sqlite3_create_collation(
         imp_dbh->db, name, SQLITE_UTF8,
-        func_sv, 
-        imp_dbh->unicode ? sqlite_db_collation_dispatcher_utf8 
+        func_sv,
+        imp_dbh->unicode ? sqlite_db_collation_dispatcher_utf8
                          : sqlite_db_collation_dispatcher
       );
 
@@ -1905,13 +1652,18 @@ sqlite_db_collation_needed(pTHX_ SV *dbh, SV *callback)
 {
     D_imp_dbh(dbh);
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to see if collation is needed on inactive database handle");
+        return;
+    }
+
     croak_if_db_is_null();
 
     /* remember the callback within the dbh */
     sv_setsv(imp_dbh->collation_needed_callback, callback);
 
     /* Register the func within sqlite3 */
-    (void) sqlite3_collation_needed( imp_dbh->db, 
+    (void) sqlite3_collation_needed( imp_dbh->db,
                                      (void*) (SvOK(callback) ? dbh : NULL),
                                      sqlite_db_collation_needed_dispatcher );
 }
@@ -1947,6 +1699,11 @@ sqlite_db_progress_handler(pTHX_ SV *dbh, int n_opcodes, SV *handler)
 {
     D_imp_dbh(dbh);
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to set progress handler on inactive database handle");
+        return FALSE;
+    }
+
     croak_if_db_is_null();
 
     if (!SvOK(handler)) {
@@ -1960,7 +1717,7 @@ sqlite_db_progress_handler(pTHX_ SV *dbh, int n_opcodes, SV *handler)
         av_push( imp_dbh->functions, handler_sv );
 
         /* Register the func within sqlite3 */
-        sqlite3_progress_handler( imp_dbh->db, n_opcodes, 
+        sqlite3_progress_handler( imp_dbh->db, n_opcodes,
                                   sqlite_db_generic_callback_dispatcher,
                                   handler_sv );
     }
@@ -1973,6 +1730,11 @@ sqlite_db_commit_hook(pTHX_ SV *dbh, SV *hook)
     D_imp_dbh(dbh);
     void *retval;
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to set commit hook on inactive database handle");
+        return &PL_sv_undef;
+    }
+
     croak_if_db_is_null();
 
     if (!SvOK(hook)) {
@@ -1986,7 +1748,7 @@ sqlite_db_commit_hook(pTHX_ SV *dbh, SV *hook)
         av_push( imp_dbh->functions, hook_sv );
 
         /* Register the hook within sqlite3 */
-        retval = sqlite3_commit_hook( imp_dbh->db, 
+        retval = sqlite3_commit_hook( imp_dbh->db,
                                       sqlite_db_generic_callback_dispatcher,
                                       hook_sv );
     }
@@ -2000,6 +1762,11 @@ sqlite_db_rollback_hook(pTHX_ SV *dbh, SV *hook)
     D_imp_dbh(dbh);
     void *retval;
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to set rollback hook on inactive database handle");
+        return &PL_sv_undef;
+    }
+
     croak_if_db_is_null();
 
     if (!SvOK(hook)) {
@@ -2013,7 +1780,7 @@ sqlite_db_rollback_hook(pTHX_ SV *dbh, SV *hook)
         av_push( imp_dbh->functions, hook_sv );
 
         /* Register the hook within sqlite3 */
-        retval = sqlite3_rollback_hook( imp_dbh->db, 
+        retval = sqlite3_rollback_hook( imp_dbh->db,
                                         (void(*)(void *))
                                         sqlite_db_generic_callback_dispatcher,
                                         hook_sv );
@@ -2023,7 +1790,7 @@ sqlite_db_rollback_hook(pTHX_ SV *dbh, SV *hook)
 }
 
 void
-sqlite_db_update_dispatcher( void *callback, int op, 
+sqlite_db_update_dispatcher( void *callback, int op,
                              char const *database, char const *table,
                              sqlite3_int64 rowid )
 {
@@ -2054,6 +1821,11 @@ sqlite_db_update_hook(pTHX_ SV *dbh, SV *hook)
     D_imp_dbh(dbh);
     void *retval;
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to set update hook on inactive database handle");
+        return &PL_sv_undef;
+    }
+
     croak_if_db_is_null();
 
     if (!SvOK(hook)) {
@@ -2067,7 +1839,7 @@ sqlite_db_update_hook(pTHX_ SV *dbh, SV *hook)
         av_push( imp_dbh->functions, hook_sv );
 
         /* Register the hook within sqlite3 */
-        retval = sqlite3_update_hook( imp_dbh->db, 
+        retval = sqlite3_update_hook( imp_dbh->db,
                                       sqlite_db_update_dispatcher,
                                       hook_sv );
     }
@@ -2125,6 +1897,11 @@ sqlite_db_set_authorizer(pTHX_ SV *dbh, SV *authorizer)
     D_imp_dbh(dbh);
     int retval;
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to set authorizer on inactive database handle");
+        return FALSE;
+    }
+
     croak_if_db_is_null();
 
     if (!SvOK(authorizer)) {
@@ -2138,7 +1915,7 @@ sqlite_db_set_authorizer(pTHX_ SV *dbh, SV *authorizer)
         av_push( imp_dbh->functions, authorizer_sv );
 
         /* Register the hook within sqlite3 */
-        retval = sqlite3_set_authorizer( imp_dbh->db, 
+        retval = sqlite3_set_authorizer( imp_dbh->db,
                                          sqlite_db_authorizer_dispatcher,
                                          authorizer_sv );
     }
@@ -2161,6 +1938,11 @@ sqlite_db_backup_from_file(pTHX_ SV *dbh, char *filename)
     sqlite3 *pFrom;
     sqlite3_backup *pBackup;
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to backup from file on inactive database handle");
+        return FALSE;
+    }
+
     croak_if_db_is_null();
 
     rc = sqlite_open(filename, &pFrom);
@@ -2204,6 +1986,11 @@ sqlite_db_backup_to_file(pTHX_ SV *dbh, char *filename)
     sqlite3 *pTo;
     sqlite3_backup *pBackup;
 
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to backup to file on inactive database handle");
+        return FALSE;
+    }
+
     croak_if_db_is_null();
 
     rc = sqlite_open(filename, &pTo);
@@ -2232,4 +2019,288 @@ sqlite_db_backup_to_file(pTHX_ SV *dbh, char *filename)
 #endif
 }
 
+typedef struct perl_tokenizer {
+    sqlite3_tokenizer base;
+    SV *coderef;                 /* the perl tokenizer is a coderef that takes
+                                    a string and returns a cursor coderef */
+} perl_tokenizer;
+
+typedef struct perl_tokenizer_cursor {
+    sqlite3_tokenizer_cursor base;
+    SV *coderef;                 /* ref to the closure that returns terms */
+    char *pToken;                /* storage for a copy of the last token */
+    int nTokenAllocated;         /* space allocated to pToken buffer */
+
+    /* members below are only used if the input string is in utf8 */
+    const char *pInput;          /* input we are tokenizing */
+    const char *lastByteOffset;  /* offset into pInput */
+    int lastCharOffset;          /* char offset corresponding to lastByteOffset */
+} perl_tokenizer_cursor;
+
+/*
+** Create a new tokenizer instance.
+** Will be called whenever a FTS3 table is created with
+**   CREATE .. USING fts3( ... , tokenize=perl qualified::function::name)
+** where qualified::function::name is a fully qualified perl function
+*/
+static int perl_tokenizer_Create(
+    int argc, const char * const *argv,
+    sqlite3_tokenizer **ppTokenizer
+){
+    dTHX;
+    dSP;
+    int n_retval;
+    SV *retval;
+
+    perl_tokenizer *t;
+    t = (perl_tokenizer *) sqlite3_malloc(sizeof(*t));
+    if( t==NULL ) return SQLITE_NOMEM;
+    memset(t, 0, sizeof(*t));
+
+    ENTER;
+    SAVETMPS;
+
+    /* call the qualified::function::name */
+    PUSHMARK(SP);
+    PUTBACK;
+    n_retval = call_pv(argv[0], G_SCALAR);
+    SPAGAIN;
+
+    /* store a copy of the returned coderef into the tokenizer structure */
+    if (n_retval != 1) {
+        warn("tokenizer_Create returned %d arguments", n_retval);
+    }
+    retval = POPs;
+    t->coderef   = newSVsv(retval);
+    *ppTokenizer = &t->base;
+
+    PUTBACK;
+    FREETMPS;
+    LEAVE;
+
+    return SQLITE_OK;
+}
+
+/*
+** Destroy a tokenizer
+*/
+static int perl_tokenizer_Destroy(sqlite3_tokenizer *pTokenizer){
+    dTHX;
+    perl_tokenizer *t = (perl_tokenizer *) pTokenizer;
+    sv_free(t->coderef);
+    sqlite3_free(t);
+    return SQLITE_OK;
+}
+
+/*
+** Prepare to begin tokenizing a particular string.  The input
+** string to be tokenized is supposed to be pInput[0..nBytes-1] ..
+** except that nBytes passed by fts3 is -1 (don't know why) !
+** This is passed to the tokenizer instance, which then returns a
+** closure implementing the cursor (so the cursor is again a coderef).
+*/
+static int perl_tokenizer_Open(
+    sqlite3_tokenizer *pTokenizer,       /* Tokenizer object */
+    const char *pInput, int nBytes,      /* Input buffer */
+    sqlite3_tokenizer_cursor **ppCursor  /* OUT: Created tokenizer cursor */
+){
+    dTHX;
+    dSP;
+    U32 flags;
+    SV *perl_string;
+    int n_retval;
+
+    perl_tokenizer *t = (perl_tokenizer *)pTokenizer;
+
+    /* allocate and initialize the cursor struct */
+    perl_tokenizer_cursor *c;
+    c = (perl_tokenizer_cursor *) sqlite3_malloc(sizeof(*c));
+    memset(c, 0, sizeof(*c));
+    *ppCursor = &c->base;
+
+    /* flags for creating the Perl SV containing the input string */
+    flags = SVs_TEMP; /* will call sv_2mortal */
+
+    /* special handling if working with utf8 strings */
+    if (last_executed_dbh->unicode) { /* global var ... no better way ! */
+
+        /* data to keep track of byte offsets */
+        c->lastByteOffset = c->pInput = pInput;
+        c->lastCharOffset = 0;
+
+        /* string passed to Perl needs to be flagged as utf8 */
+        flags |= SVf_UTF8;
+    }
+
+    ENTER;
+    SAVETMPS;
+
+    /* build a Perl copy of the input string */
+    if (nBytes < 0) { /* we get -1 from fts3. Don't know why ! */
+        nBytes = strlen(pInput);
+    }
+    perl_string = newSVpvn_flags(pInput, nBytes, flags);
+
+    /* call the tokenizer coderef */
+    PUSHMARK(SP);
+    XPUSHs(perl_string);
+    PUTBACK;
+    n_retval = call_sv(t->coderef, G_SCALAR);
+    SPAGAIN;
+
+    /* store the cursor coderef returned by the tokenizer */
+    if (n_retval != 1) {
+        warn("tokenizer returned %d arguments", n_retval);
+    }
+    c->coderef = newSVsv(POPs);
+
+    PUTBACK;
+    FREETMPS;
+    LEAVE;
+    return SQLITE_OK;
+}
+
+/*
+** Close a tokenization cursor previously opened by a call to
+** perl_tokenizer_Open() above.
+*/
+static int perl_tokenizer_Close(sqlite3_tokenizer_cursor *pCursor){
+    perl_tokenizer_cursor *c = (perl_tokenizer_cursor *) pCursor;
+
+    dTHX;
+    sv_free(c->coderef);
+    sqlite3_free(c);
+    return SQLITE_OK;
+}
+
+/*
+** Extract the next token from a tokenization cursor.  The cursor must
+** have been opened by a prior call to perl_tokenizer_Open().
+*/
+static int perl_tokenizer_Next(
+    sqlite3_tokenizer_cursor *pCursor,  /* Cursor returned by perl_tokenizer_Open */
+    const char **ppToken,               /* OUT: *ppToken is the token text */
+    int *pnBytes,                       /* OUT: Number of bytes in token */
+    int *piStartOffset,                 /* OUT: Starting offset of token */
+    int *piEndOffset,                   /* OUT: Ending offset of token */
+    int *piPosition                     /* OUT: Position integer of token */
+){
+    perl_tokenizer_cursor *c = (perl_tokenizer_cursor *) pCursor;
+    int result;
+    int n_retval;
+    STRLEN n_a;
+    char *token;
+    char *byteOffset;
+    I32 hop;
+
+    dTHX;
+    dSP;
+
+    ENTER;
+    SAVETMPS;
+
+    /* call the cursor */
+    PUSHMARK(SP);
+    PUTBACK;
+    n_retval = call_sv(c->coderef, G_ARRAY);
+    SPAGAIN;
+
+    /* if we get back an empty list, there is no more token */
+    if (n_retval == 0) {
+        result = SQLITE_DONE;
+    }
+    /* otherwise, get token details from the return list */
+    else {
+        if (n_retval != 5) {
+            warn("tokenizer cursor returned %d arguments", n_retval);
+        }
+        *piPosition    = POPi;
+        *piEndOffset   = POPi;
+        *piStartOffset = POPi;
+        *pnBytes       = POPi;
+        token          = POPpx;
+
+        if (c->pInput) { /* if working with utf8 data */
+
+            /* recompute *pnBytes in bytes, not in chars */
+            *pnBytes = strlen(token);
+
+            /* recompute start/end offsets in bytes, not in chars */
+            hop            = *piStartOffset - c->lastCharOffset;
+            byteOffset     = utf8_hop((U8*)c->lastByteOffset, hop);
+            hop            = *piEndOffset - *piStartOffset;
+            *piStartOffset = byteOffset - c->pInput;
+            byteOffset     = utf8_hop(byteOffset, hop);
+            *piEndOffset   = byteOffset - c->pInput;
+
+            /* remember where we are for next round */
+            c->lastCharOffset = *piEndOffset,
+            c->lastByteOffset = byteOffset;
+        }
+
+        /* make sure we have enough storage for copying the token */
+        if (*pnBytes > c->nTokenAllocated ){
+            char *pNew;
+            c->nTokenAllocated = *pnBytes + 20;
+            pNew = sqlite3_realloc(c->pToken, c->nTokenAllocated);
+            if( !pNew ) return SQLITE_NOMEM;
+            c->pToken = pNew;
+        }
+
+        /* need to copy the token into the C cursor before perl frees that
+           memory */
+        memcpy(c->pToken, token, *pnBytes);
+        *ppToken  = c->pToken;
+
+        result = SQLITE_OK;
+    }
+
+    PUTBACK;
+    FREETMPS;
+    LEAVE;
+
+    return result;
+}
+
+/*
+** The set of routines that implement the perl tokenizer
+*/
+sqlite3_tokenizer_module perl_tokenizer_Module = {
+    0,
+    perl_tokenizer_Create,
+    perl_tokenizer_Destroy,
+    perl_tokenizer_Open,
+    perl_tokenizer_Close,
+    perl_tokenizer_Next
+};
+
+/*
+** Register the perl tokenizer with FTS3
+*/
+int sqlite_db_register_fts3_perl_tokenizer(pTHX_ SV *dbh)
+{
+    D_imp_dbh(dbh);
+
+    int rc;
+    sqlite3_stmt *pStmt;
+    const char zSql[] = "SELECT fts3_tokenizer(?, ?)";
+    sqlite3_tokenizer_module *p = &perl_tokenizer_Module;
+
+    if (!DBIc_ACTIVE(imp_dbh)) {
+        sqlite_error(dbh, -2, "attempt to register fts3 tokenizer on inactive database handle");
+        return FALSE;
+    }
+
+    rc = sqlite3_prepare_v2(imp_dbh->db, zSql, -1, &pStmt, 0);
+    if( rc!=SQLITE_OK ){
+        return rc;
+    }
+
+    sqlite3_bind_text(pStmt, 1, "perl", -1, SQLITE_STATIC);
+    sqlite3_bind_blob(pStmt, 2, &p, sizeof(p), SQLITE_STATIC);
+    sqlite3_step(pStmt);
+
+    return sqlite3_finalize(pStmt);
+}
+
 /* end */
