@@ -10,7 +10,7 @@ use vars qw{$err $errstr $drh $sqlite_version $sqlite_version_number};
 use vars qw{%COLLATION};
 
 BEGIN {
-    $VERSION = '1.30_06';
+    $VERSION = '1.32_01';
     @ISA     = 'DynaLoader';
 
     # Initialize errors
@@ -79,9 +79,9 @@ sub connect {
     my ($drh, $dbname, $user, $auth, $attr) = @_;
 
     # Default PrintWarn to the value of $^W
-    unless ( defined $attr->{PrintWarn} ) {
-        $attr->{PrintWarn} = $^W ? 1 : 0;
-    }
+    # unless ( defined $attr->{PrintWarn} ) {
+    #    $attr->{PrintWarn} = $^W ? 1 : 0;
+    # }
 
     my $dbh = DBI::_new_dbh( $drh, {
         Name => $dbname,
@@ -91,7 +91,7 @@ sub connect {
     if ( $dbname =~ /=/ ) {
         foreach my $attrib ( split(/;/, $dbname) ) {
             my ($key, $value) = split(/=/, $attrib, 2);
-            if ( $key eq 'dbname' ) {
+            if ( $key =~ /^(?:db(?:name)?|database)$/ ) {
                 $real = $value;
             } else {
                 $attr->{$key} = $value;
@@ -136,9 +136,21 @@ sub connect {
 
     # HACK: Since PrintWarn = 0 doesn't seem to actually prevent warnings
     # in DBD::SQLite we set Warn to false if PrintWarn is false.
-    unless ( $attr->{PrintWarn} ) {
-        $attr->{Warn} = 0;
-    }
+
+    # NOTE: According to the explanation by timbunce,
+    # "Warn is meant to report on bad practices or problems with
+    # the DBI itself (hence always on by default), while PrintWarn
+    # is meant to report warnings coming from the database."
+    # That is, if you want to disable an ineffective rollback warning
+    # etc (due to bad practices), you should turn off Warn,
+    # and to silence other warnings, turn off PrintWarn.
+    # Warn and PrintWarn are independent, and turning off PrintWarn
+    # does not silence those warnings that should be controlled by
+    # Warn.
+
+    # unless ( $attr->{PrintWarn} ) {
+    #     $attr->{Warn} = 0;
+    # }
 
     return $dbh;
 }
@@ -163,6 +175,7 @@ sub install_collation {
 # (see http://www.sqlite.org/vtab.html#xfindfunction)
 sub regexp {
     use locale;
+    return if !defined $_[0] || !defined $_[1];
     return scalar($_[1] =~ $_[0]);
 }
 
